@@ -6,6 +6,55 @@ from natsort import natsorted, ns
 
 from .rml import RMLFile
 
+class RayProperties(np.ndarray):
+    """RayProperties class privides simplified interface to access results of the 
+    postprocessing of calculated rays and provides functions to read and write those results
+    into a txt file, including the header
+
+    Args:
+        input (nd.array): create RayProperties from an array. Intentende for internal use only!
+        filename (str): file name to load the data from (including the header information/row)
+    """
+    def __new__(cls,input:np.ndarray=None,/,filename=None) -> None:
+        dt_names = ['SourcePhotonFlux', 'NumberRaysSurvived', 'PercentageRaysSurvived', 'PhotonFlux', 'Bandwidth', 'HorizontalFocusFWHM', 'VerticalFocusFWHM']
+        dt_formats = [float for n in dt_names]
+        dt = np.dtype({'names':dt_names, 'formats':dt_formats})
+
+        if input is None:
+            if filename is None:
+                input_array = np.zeros(1, dtype=dt)
+            else:
+                input_array = np.genfromtxt(filename, dtype=float, delimiter='\t', names=True)
+        else:
+            input_array = input.copy()
+            input_array.dtype = dt
+
+        obj = np.asarray(input_array).view(cls)
+        return obj
+
+    def save(self,filename):
+        """Save current object into tab-separated file with header row
+
+        Args:
+            filename (_type_): _description_
+        """
+        tmp = self.copy()
+        if tmp.shape[0]>1:
+            tmp.dtype = float
+        np.savetxt(filename,tmp,delimiter='\t',header="\t".join(self.dtype.names))
+        
+
+    def concat(self,other):
+        """concatenate another instalnce of RayProperties and return a new object 
+
+        Args:
+            other (RayProperties): instance of RayProperties to be contated to this instance
+
+        Returns:
+            RayProperties: resulting RayProperties object
+        """
+        return RayProperties(np.vstack((self,other)))
+
 
 class PostProcess():
     """class to post-process the data. 
@@ -36,28 +85,42 @@ class PostProcess():
                 res.append(os.path.join(dir_path,file))
         return natsorted(res, alg=ns.IGNORECASE)
 
-    def _extract_bandwidth_fwhm(self,rays_bw:np.array):
-        """calculate the fwhm of the rays_bw.
+    def _extract_fwhm(self,rays:np.array):
+        """Calculate the fwhm of the rays.
+
+        If less than 100 rays are passed check what is the standard deviation of the array.
+        Else, make an histogram with 30 bins and check when the array falls at less than half the max
 
         Args:
-            rays_bw (np:array): the energy of the x-rays
+            rays (np:array): the energy of the x-rays
 
         Returns:
             float: fwhm
         """        
-        return 2.3555*np.std(rays_bw)
+        # if I have less than 100 rays calculate the standard deviation
+        if rays.shape[0]<100:
+            return 2*np.sqrt(2*np.log(2))*np.std(rays)
+        
+        # else actually look for the fwhm
 
-    def _extract_focus_fwhm(self,rays_pos:np.array):
-        """calculate the fwhm of rays_pos
+        # make an histogram, get back a tuple of values and bins
+        gh = np.histogram(rays, bins=30)
+        y = gh[0]
+        x_bins = gh[1]
 
-        Args:
-            rays_pos (np.array): contains positions of the x-rays
+        # take the average of each pari of bins to get the middle
+        x = (x_bins[1:] + x_bins[:-1]) / 2
 
-        Returns:
-            float: fwhm
-        """        
-        return 2.3555*np.std(rays_pos)
+        # Find the maximum y value
+        max_y = np.amax(y)  
 
+        # check where y becomes higher that max_y/2
+        xs = [x for x in range(y.shape[0]) if y[x] > max_y/2.0]
+        fwhm = x[np.amax(xs)-1]-x[np.amin(xs)-1]
+        fwhm_std = 2*np.sqrt(2*np.log(2))*np.std(rays)
+        #print(f'DEBUG:: fwhm: {fwhm} vs np.std() {fwhm_std}')
+        return fwhm
+        
     def _extract_intensity(self,rays:np.array):
         """calculate how many rays there are
 
@@ -66,31 +129,6 @@ class PostProcess():
         """        
         return(rays.shape[0])
     
-    def _save_file(self, filename:str, array:np.array, header:str=None):
-        """This function is used to save files.
-
-        Args:
-            filename (str): file name(path)
-            array (np.array): array to save 
-            header (str): header for the file
-        """        
-        if header != None:
-            np.savetxt(filename+self.format_saved_files,array, header=header)
-        else:
-            np.savetxt(filename+self.format_saved_files,array)
-
-    def _load_file(self,filepath):
-        """Load file and returns the array
-
-        Args:
-            filepath (str): the path to the file to load
-
-        Returns:
-            arr (np.array): The loaded numpy array
-        """        
-        arr = np.loadtxt(filepath)
-        return arr
-    
     def extract_nrays_from_source(self, rml_filename):
         """Extract photon flux from rml file, find source automatically
 
@@ -129,45 +167,25 @@ class PostProcess():
         filename = os.path.join(dir_path,sim_number+exported_element + '-' + exported_object+'.csv')
         with warnings.catch_warnings():
             warnings.simplefilter("ignore")
-            rays = np.loadtxt(filename, skiprows=2)
-        ray_properties = np.zeros((7,1))
+            #rays = np.loadtxt(filename, skiprows=2)
+            rays = np.genfromtxt(filename, dtype=float, delimiter='\t', names=True,skip_header=1)
+        ray_properties = RayProperties()
         if rays.shape[0]==0: # if no rays survived
             # source photon flux
-            ray_properties[0] = source_photon_flux
+            ray_properties['SourcePhotonFlux'] = source_photon_flux
             pass
-        elif rays.shape[0]==15: # if only one ray survived
-            # source photon flux
-            ray_properties[0] = source_photon_flux
-            # number of rays reaching the oe
-            ray_properties[1] = 1
-            # percentage of survived photons
-            ray_properties[2] = ray_properties[1]/source_n_rays*100
-            # photon flux reaching the oe
-            ray_properties[3] = source_photon_flux/100*ray_properties[2]
-            # bandwidth of the rays reaching the OE
-            ray_properties[4] = self._extract_bandwidth_fwhm(rays[9])
-            # horizontal focus
-            ray_properties[5] = self._extract_focus_fwhm(rays[3])
-            # vertical focus
-            ray_properties[6] = self._extract_focus_fwhm(rays[4])
         else:
-            # source photon flux
-            ray_properties[0] = source_photon_flux
-            # number of rays reaching the oe
-            ray_properties[1] = self._extract_intensity(rays)
-            # percentage of survived photons
-            ray_properties[2] = ray_properties[1]/source_n_rays*100
-            # photon flux reaching the oe
-            ray_properties[3] = source_photon_flux/100*ray_properties[2]
-            # bandwidth of the rays reaching the oe
-            ray_properties[4] = self._extract_bandwidth_fwhm(rays[:,9])
-            # horizontal focus
-            ray_properties[5] = self._extract_focus_fwhm(rays[:,3])
-            # vertical focus
-            ray_properties[6] = self._extract_focus_fwhm(rays[:,4])
+            ray_properties['SourcePhotonFlux'] = source_photon_flux
+            ray_properties['NumberRaysSurvived'] = self._extract_intensity(rays)
+            ray_properties['PercentageRaysSurvived'] = ray_properties['NumberRaysSurvived']/source_n_rays*100
+            ray_properties['PhotonFlux'] = source_photon_flux/100*ray_properties['PercentageRaysSurvived']
+            ray_properties['Bandwidth'] = self._extract_fwhm(rays[f'{exported_element}_EN'])
+            ray_properties['HorizontalFocusFWHM'] = self._extract_fwhm(rays[f'{exported_element}_OX'])
+            ray_properties['VerticalFocusFWHM'] = self._extract_fwhm(rays[f'{exported_element}_OY'])
         
-        new_filename = os.path.join(dir_path, sim_number+exported_element+'_analyzed_rays')
-        self._save_file(new_filename, ray_properties)
+        new_filename = os.path.join(dir_path, sim_number+exported_element+'_analyzed_rays.dat')
+        #self._save_file(new_filename, ray_properties)
+        ray_properties.save(new_filename)
         return 
 
     def cleanup(self,dir_path:str=None, repeat:int=1, exp_elements:list=None):
@@ -190,21 +208,20 @@ class PostProcess():
                 files = self._list_files(dir_path_round, d[0]+"_analyzed_rays"+self.format_saved_files)
                 for f_ind, f in enumerate(files):
                     if r == 0 and f_ind==0:
-                        analyzed_rays = self._load_file(f)
-                        analyzed_rays = np.reshape(analyzed_rays,(1,analyzed_rays.shape[0]))
+                        analyzed_rays = RayProperties(filename=f)
                     elif r==0 and f_ind!=0:
-                        tmp=self._load_file(f)
-                        tmp = np.reshape(tmp,(1,tmp.shape[0]))
-                        analyzed_rays = np.concatenate((analyzed_rays, tmp), axis=0)
+                        tmp = RayProperties(filename=f)
+                        analyzed_rays = analyzed_rays.concat(tmp)
                     elif r>=1:
-                        tmp=self._load_file(f)
-                        tmp=tmp.reshape((tmp.shape[0]))
-                        analyzed_rays[f_ind,:] += tmp
+                        tmp = RayProperties(filename=f)
+                        for n in analyzed_rays.dtype.names: 
+                            analyzed_rays[n][f_ind] += tmp[n]
                     else:
                         pass
             fn = os.path.join(dir_path, d[0])
-            analyzed_rays = analyzed_rays/repeat
-            self._save_file(fn,analyzed_rays,header=header)
+            for n in analyzed_rays.dtype.names:
+                analyzed_rays[n] /= repeat
+            analyzed_rays.save(f"{fn}.dat")
 
 class PostProcessAnalyzed():
     """class to analyze the data exported by RAY-UI
@@ -262,7 +279,7 @@ class PostProcessAnalyzed():
         if source == 'Dipole':
             flux = np.array(flux)
             flux_dipole = np.array(flux_dipole)
-            return flux*scale_factor, flux_percent*scale_factor, flux_dipole*scale_factor
+            return flux*scale_factor, flux_percent, flux_dipole*scale_factor
         return flux_percent*scale_factor
 
     def retrieve_bw_and_focusSize(self,folder_name:str,oe:str,nsimulations:int,rounds:int):
